/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.listener.outgoing;

import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.JBIException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.Source;

import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.OMText;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.addressing.EndpointReference;
import org.apache.axis2.client.Options;
import org.apache.axis2.client.ServiceClient;
import org.apache.axis2.transport.http.HTTPConstants;
import org.apache.axis2.transport.http.HttpTransportProperties;
import org.ow2.petals.binding.soap.SoapComponentContext;
import org.ow2.petals.binding.soap.listener.SUPropertiesHelper;
import org.ow2.petals.binding.soap.util.AxiomUtils;
import org.ow2.petals.binding.soap.util.Marshaller;
import org.ow2.petals.binding.soap.util.StaxUtils;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.performance.PerformanceNotification;

import static org.ow2.petals.binding.soap.Constants.ServiceUnit.ADDRESS;

/**
 * An external web service dispatcher. This dispatcher send the JBI message to
 * an external web service. The service URL is specified in the PEtALS
 * extensions.
 * 
 * @author chamerling - eBM WebSourcing
 * 
 */
public class WSCaller extends AbstractJBIDispatcher {

    /**
     * The JBI component instance
     */
    private AbstractComponent component = null;

    /**
     * Creates a new instance of {@link WSCaller}
     * 
     * @param soapContext
     * @param logger
     */
    public WSCaller(final SoapComponentContext soapContext, final Logger logger,
            final AbstractComponent component) {
        super(soapContext, logger);
        this.component = component;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.binding.soap.listener.outgoing.JBIListenerDispatcher #
     *      process(org.objectweb.petals.component.common.util.MessageExchangeWrapper ,
     *      org.objectweb.petals.common.descriptor.Extensions)
     */
    public void process(final Exchange exchange, final ConfigurationExtensions cdkExtensions) {
        final String address = cdkExtensions.get(ADDRESS);

        this.logger.fine("Calling external web service : " + address);

        // Get the incoming Normalized message
        final NormalizedMessage in = exchange.getInMessage();
        if (in == null) {
            exchange.setError(new JBIException(
            "Message exchange must handle an In normalized message"));
        } else {

            // Get the outgoing Normalized message
            final NormalizedMessage out = exchange.getOutMessage();

            // Get operation
            final QName jbiOperation = exchange.getOperation();
            final QName soapAction = this.getSOAPAction(jbiOperation, cdkExtensions);

            // create Options for the stub
            final Options options = this.createOptions(address, soapAction, cdkExtensions);

            // create service client used to invoke WS
            try {
                final ServiceClient serviceClient = Axis2Utils.createServiceClient(exchange.getExchangePattern(),
                        soapAction);
                serviceClient.setOptions(options);

                // create the in body
                final OMElement inBodyElement = this.createSOAPBodyContent(in, exchange
                        .getOperation(), serviceClient, cdkExtensions);

                if (this.logger.isLoggable(Level.FINE)) {
                    // The condition is splitted for performance reasons
                    if (inBodyElement != null) {
                        this.logger.log(Level.FINE, "OUTGOING Payload : " + inBodyElement.toString());
                    }
                }

                // engage the Axis2 modules
                this.engageModules(address, serviceClient);

                if (exchange.isInOnlyPattern()) {
                    // send as InOnly message
                    serviceClient.fireAndForget(soapAction, inBodyElement);

                } else if (exchange.isRobustInOnlyPattern()) {
                    // send as RobustInOnly message
                    // If the robust in only exchange throws a Fault
                    // It will be catch by the component-common framework
                    serviceClient.sendRobust(soapAction, inBodyElement);

                } else if (exchange.isInOptionalOutPattern() || exchange.isInOutPattern()) {
                    // Performance notification creation
                    PerformanceNotification performanceNotification = null;
                    performanceNotification = this.component
                    .createPerformanceNotificationSendingToOutside(address, exchange);

                    // send as In(optional)Out message
                    OMElement outBodyElement = null;
                    try {
                        outBodyElement = serviceClient.sendReceive(soapAction, inBodyElement);
                    } catch (final AxisFault e) {
                        // only catched to send the performance notification.
                        // TODO : Extend to all MEPs
                        throw e;
                    } finally {
                        // Performance notification sending
                        this.component.sendPerformanceNotification(performanceNotification);
                    }

                    // if msg exchange required a response, set it
                    if (outBodyElement != null) {
                        // For performance reasons, the AXIOM's cache is not
                        // used to optimize the namespace writing. The service
                        // provider must be in charge of this optimization.
                        if (this.logger.isLoggable(Level.FINE)) {
                            this.logger.log(Level.FINE, "RESPONSE Payload : " + outBodyElement.toStringWithConsume());
                        }

                        out.setContent(AxiomUtils.createSource(outBodyElement));
                        Marshaller.copyAttachments(outBodyElement, out);
                    }
                    else {
                        this.logger.log(Level.FINE, "RESPONSE Payload : No response.");
                    }
                } else {
                    throw new MessagingException("WSCaller - MEP not recognized : "
                            + exchange.getPattern().toString());
                }
            } catch (final Exception e) {
                this.logger.warning("Catch an exception on the WS invocation : " + e.getMessage());

                // The exception is an AxisFault (probably a SOAP Fault)
                if (e instanceof AxisFault) {
                    this.handleSOAPFault(exchange, (AxisFault) e);
                } else {
                    this.handleException(exchange, e);
                }
            }
        }
    }

    /**
     * Set the Exception in the JBI Message as Fault
     * 
     * @param exchange
     * @param e
     */
    private void handleException(final Exchange exchange, final Exception e) {
        exchange.setError(new MessagingException("BC-SOAP Exception => " + e.getMessage(),e));
    }

    /**
     * Handle a SOAP Fault. Set it in the JBI message as Fault.
     * 
     * @param exchange
     * @param fault
     */
    private void handleSOAPFault(final Exchange exchange, final AxisFault fault) {
        if ((fault.getFaultMessageContext() != null)
                && (fault.getFaultMessageContext().getEnvelope() != null)) {
            try {
                final Source sourceFault = Marshaller.createSource(fault
                        .getFaultMessageContext().getEnvelope(), false);
                final Fault jbiFault = exchange.createFault();
                jbiFault.setContent(sourceFault);
                exchange.setFault(jbiFault);
            } catch (final Exception e1) {
                this.logger.log(Level.SEVERE, "Can't return fault to consumer", e1);
            }
        } else {
            this.handleException(exchange, fault);
        }
    }

    /**
     * Create the SOAP options of the outgoing SOAP message
     * 
     * @param address
     * @param operation
     * @param extensions
     * @return
     */
    protected Options createOptions(final String address, final QName soapAction,
            final ConfigurationExtensions extensions) {

        final Options options = new Options();

        // set destination address
        options.setTo(new EndpointReference(address));

        // set the soap action
        if (soapAction != null) {
            options.setAction(soapAction.toString());
        }

        // set the timeout
        final long timeout = SUPropertiesHelper.retrieveTimeout(extensions);
        if (timeout != -1L) {
            options.setTimeOutInMilliSeconds(timeout);
        }

        // Get the soapEnvelopeNamespaceURI version to use - optional, default
        // is 1.1 version
        final String soapEnvelopeNamespaceURI = this.retrieveSOAPEnvelopeNamespaceURI(extensions);
        options.setSoapVersionURI(soapEnvelopeNamespaceURI);

        // set transport
        options.setTransportInProtocol(Constants.TRANSPORT_HTTP);

        // get proxy settings if they are defined in the extensions
        final HttpTransportProperties.ProxyProperties proxyProperties = this
        .retrieveProxySettings(extensions);
        if (proxyProperties != null) {
            options.setProperty(HTTPConstants.PROXY, proxyProperties);
        }

        options.setProperty(org.apache.axis2.transport.http.HTTPConstants.CHUNKED, this
                .retrieveChunkedMode(extensions));

        options.setCallTransportCleanup(this.retrieveCleanupTransport(extensions));
        return options;
    }

    /**
     * Create the SOAPBody content. The JBI content is used as root element. All
     * the JBI attachments are added as root element children.
     * 
     * @param body
     * @return the soap body content or null if {@link NormalizedMessage} is
     *  empty.
     * @throws XMLStreamException
     */
    @SuppressWarnings("unchecked")
    protected OMElement createSOAPBodyContent(final NormalizedMessage nm, final QName operation,
            final ServiceClient client, final ConfigurationExtensions extensions)
    throws XMLStreamException {

        OMElement document = null;
        final Source src = nm.getContent();

        if (src != null) {
            // create the root element
            final OMFactory fac = OMAbstractFactory.getOMFactory();
            document = this.buildOMElementFromSource(src, operation, extensions, fac);

            // add attachments if any
            if (nm.getAttachmentNames().size() > 0) {
                // enable MTOM
                client.getOptions().setProperty(Constants.Configuration.ENABLE_MTOM,
                        Constants.VALUE_TRUE);

                // Add JBI attachments to the document element
                final Set<String> names = nm.getAttachmentNames();
                for (final String key : names) {
                    final DataHandler attachment = nm.getAttachment(key);
                    final OMText attach = fac.createOMText(attachment, true);
                    document.addChild(attach);
                }
            }
        }
        return document;
    }

    /**
     * TODO Enable the client to contact a service over an SSL connection. The
     * required values are in the PEtALS extensions if the SU has been
     * parametrized to enable SSL.
     * 
     * @param extensions
     */
    protected void enableSSLSupport(final ConfigurationExtensions extensions) {
        final String trustStore = extensions.get("ssl-trustStore");
        final String trustStorePassword = extensions.get("ssl-trustStorePassword");
        final String keyStore = extensions.get("ssl-keyStore");
        final String keyStorePassword = extensions.get("ssl-keyStorePassword");

        if ((trustStore != null) && (trustStorePassword != null) && (keyStore != null)
                && (keyStorePassword != null)) {
            // TODO
        }
    }

    /**
     * Get the operation for the output SOAP message. Get the operation defined
     * in the SU first, if null get the JBI operation.
     * 
     * @param jbiOperation
     * @param extensions
     * @return
     */
    protected QName getSOAPAction(final QName jbiOperation, final ConfigurationExtensions extensions) {
        QName soapAction = this.retrieveDefaultSOAPAction(extensions);
        if (soapAction == null) {
            soapAction = jbiOperation;
        }
        return soapAction;
    }

    /**
     * Build the {@link OMElement} from the {@link Source}. The SOAP operation
     * is added as root element if it is configured in the petals extensions.
     * 
     * @param body
     * @param operation
     * @param extensions
     * @return
     */
    protected OMElement buildOMElementFromSource(final Source body, final QName operation,
            final ConfigurationExtensions extensions, final OMFactory factory)
    throws XMLStreamException {
        final XMLStreamReader parser = StaxUtils.createXMLStreamReader(body);
        final StAXOMBuilder builder = new StAXOMBuilder(parser);
        OMElement rootElement = builder.getDocumentElement();

        if (this.retrieveAddRoot(extensions)) {
            final OMElement newRootElement = factory.createOMElement(operation);
            newRootElement.addChild(rootElement);
            rootElement = newRootElement;
        }
        return rootElement;
    }

    /**
     * Engage the modules for the service client. The modules are available on
     * the service unit listener since they have been defined during SU
     * deployment.
     * 
     * @param serviceClient
     * @throws AxisFault
     */
    private void engageModules(final String address, final ServiceClient serviceClient)
    throws AxisFault {
        final List<String> modules = this.soapContext.getModules(address);
        if (modules == null) {
            return;
        }

        for (final String name : modules) {
            this.logger.fine("Engaging module " + name);
            serviceClient.engageModule(name);
        }
    }

    public String getDispatcherType() {
        return "SOAP";
    }
}
